"""Totally typed, plausibly practical, and remarkably random utilities—for me, and maybe for you too."""

from kajihs_utils.core import bisect_predicate, get_first, is_sorted

__all__ = ["bisect_predicate", "get_first", "is_sorted"]
