"""kajihs-utils metadata."""

__module_name__ = "kajihs_utils"  # Not dynamic
__version__ = "0.10.1"

__authors__ = ["Julian Paquerot"]
__author_emails__ = ["itskajih@gmail.com"]

# === URLs ===
# https://packaging.python.org/en/latest/specifications/well-known-project-urls/#well-known-labels
__repo_url__ = "https://github.com/Kajiih/kajihs-utils"
__homepage_url__ = "https://github.com/Kajiih/kajihs-utils"
__issues_url__ = f"{__repo_url__}/issues"
__changelog_url__ = f"{__repo_url__}/blob/main/CHANGELOG.md"
__documentation_url__ = "https://Kajiih.github.io/kajihs-utils/"
